"use strict";

import { ACCESSIBLE } from "./enums.js";
class SecureStorage {
  functionCache = {};
  nativeInstance = globalThis.__SecureStorage;
  getFunctionFromCache(functionName) {
    if (!this.functionCache[functionName]) {
      this.functionCache[functionName] = this.nativeInstance?.[functionName];
    }
    return this.functionCache[functionName];
  }

  /**
   * Stores an item in the secure storage.
   * @param key - The key of the item.
   * @param value - The value of the item.
   * @param accessible - (Optional) The accessibility level of the item. Default is ACCESSIBLE.WHEN_UNLOCKED.
   * @returns A promise that resolves to a boolean indicating the success of the operation.
   */
  setItem = async (key, value, accessible = ACCESSIBLE.WHEN_UNLOCKED) => {
    const func = this.getFunctionFromCache("setItem");
    const result = await func(key, value, accessible);
    return result;
  };

  /**
   * Retrieves an item from the secure storage.
   * @param key - The key of the item to retrieve.
   * @returns A promise that resolves to the value of the item.
   */
  getItem = async key => {
    const func = this.getFunctionFromCache("getItem");
    const result = await func(key);
    return result;
  };

  /**
   * Clears all items from the secure storage.
   * @returns A promise that resolves when the operation is complete.
   */
  clearStorage = async () => {
    const func = this.getFunctionFromCache("clearStorage");
    await func();
  };

  /**
   * Stores multiple items in the secure storage.
   * @param items - An array of items to set.
   * @returns A promise that resolves to a boolean indicating the success of the operation.
   */
  setItems = async items => {
    const func = this.getFunctionFromCache("setItems");
    const result = func(items);
    return result;
  };

  /**
   * Retrieves all keys from the secure storage.
   * @returns A promise that resolves to an array of all keys.
   */
  getAllKeys = async () => {
    const func = this.getFunctionFromCache("getAllKeys");
    const result = await func();
    return JSON.parse(result);
  };

  /**
   * Retrieves all items from the secure storage.
   * @returns A promise that resolves to an array of all items.
   */
  getAllItems = async () => {
    const func = this.getFunctionFromCache("getAllItems");
    const result = await func();
    return JSON.parse(result);
  };

  /**
   * Removes an item from the secure storage.
   * @param key - The key of the item to remove.
   * @returns A promise that resolves to a boolean indicating the success of the operation.
   */
  removeItem = async key => {
    const func = this.getFunctionFromCache("removeItem");
    const result = await func(key);
    return result;
  };

  /**
   * Checks if an item exists in the secure storage.
   * @param key - The key of the item to check.
   * @returns A promise that resolves to a boolean indicating whether the item exists.
   */
  hasItem = async key => {
    const func = this.getFunctionFromCache("hasItem");
    const result = await func(key);
    return result;
  };
  setItemSync = (key, value, accessible = ACCESSIBLE.WHEN_UNLOCKED) => {
    const func = this.getFunctionFromCache("setItemSync");
    return func(key, value, accessible);
  };
  getItemSync = key => {
    const func = this.getFunctionFromCache("getItemSync");
    return func(key);
  };
  removeItemSync = key => {
    const func = this.getFunctionFromCache("removeItemSync");
    return func(key);
  };
  hasItemSync = key => {
    const func = this.getFunctionFromCache("hasItemSync");
    return func(key);
  };
}
export default SecureStorage;
//# sourceMappingURL=SecureStorage.js.map