import { ACCESSIBLE } from "./enums";
import type { ISecureStorage, SecureStorageItem, StoredSecureStorageItem } from "./types";
declare class SecureStorage implements ISecureStorage {
    private functionCache;
    private nativeInstance;
    private getFunctionFromCache;
    /**
     * Stores an item in the secure storage.
     * @param key - The key of the item.
     * @param value - The value of the item.
     * @param accessible - (Optional) The accessibility level of the item. Default is ACCESSIBLE.WHEN_UNLOCKED.
     * @returns A promise that resolves to a boolean indicating the success of the operation.
     */
    setItem: (key: string, value: string, accessible?: ACCESSIBLE) => Promise<boolean>;
    /**
     * Retrieves an item from the secure storage.
     * @param key - The key of the item to retrieve.
     * @returns A promise that resolves to the value of the item.
     */
    getItem: (key: string) => Promise<string>;
    /**
     * Clears all items from the secure storage.
     * @returns A promise that resolves when the operation is complete.
     */
    clearStorage: () => Promise<void>;
    /**
     * Stores multiple items in the secure storage.
     * @param items - An array of items to set.
     * @returns A promise that resolves to a boolean indicating the success of the operation.
     */
    setItems: (items: SecureStorageItem[]) => Promise<boolean>;
    /**
     * Retrieves all keys from the secure storage.
     * @returns A promise that resolves to an array of all keys.
     */
    getAllKeys: () => Promise<StoredSecureStorageItem>;
    /**
     * Retrieves all items from the secure storage.
     * @returns A promise that resolves to an array of all items.
     */
    getAllItems: () => Promise<StoredSecureStorageItem>;
    /**
     * Removes an item from the secure storage.
     * @param key - The key of the item to remove.
     * @returns A promise that resolves to a boolean indicating the success of the operation.
     */
    removeItem: (key: string) => Promise<boolean>;
    /**
     * Checks if an item exists in the secure storage.
     * @param key - The key of the item to check.
     * @returns A promise that resolves to a boolean indicating whether the item exists.
     */
    hasItem: (key: string) => Promise<boolean>;
    setItemSync: (key: string, value: string, accessible?: ACCESSIBLE) => boolean;
    getItemSync: (key: string) => string | null;
    removeItemSync: (key: string) => boolean;
    hasItemSync: (key: string) => boolean;
}
export default SecureStorage;
//# sourceMappingURL=SecureStorage.d.ts.map